#! /usr/bin/env python

# standard library
import os
import unittest

# 3rd party
import yaml
from pytransifex.exceptions import PyTransifexException

# project
from qgispluginci.parameters import Parameters
from qgispluginci.translation import Translation

# Tests
from .utils import can_skip_test


class TestTranslation(unittest.TestCase):
    def setUp(self):
        with open(".qgis-plugin-ci") as f:
            arg_dict = yaml.safe_load(f)
        self.parameters = Parameters(arg_dict)
        self.transifex_token = os.getenv("transifex_token")
        self.assertIsNotNone(self.transifex_token)
        self.t = Translation(self.parameters, transifex_token=self.transifex_token)

    def tearDown(self):
        try:
            self.t._t.delete_project(self.parameters.project_slug)
        except PyTransifexException:
            pass
        try:
            self.t._t.delete_team("{}-team".format(self.parameters.project_slug))
        except PyTransifexException:
            pass

    @unittest.skipIf(can_skip_test(), "Missing transifex_token")
    def test_creation(self):
        self.t = Translation(self.parameters, transifex_token=self.transifex_token)
        self.tearDown()
        self.t = Translation(self.parameters, transifex_token=self.transifex_token)

    @unittest.skipIf(can_skip_test(), "Missing transifex_token")
    def test_pull(self):
        self.t.pull()
        self.t.compile_strings()

    @unittest.skipIf(can_skip_test(), "Missing transifex_token")
    def test_push(self):
        self.t.update_strings()
        self.t.push()


if __name__ == "__main__":
    unittest.main()



import unittest

from qgispluginci.utils import parse_tag
from qgispluginci.version_note import VersionNote


class TestUtils(unittest.TestCase):
    def test_version_note_from_tag(self):
        """Test to parse a tag and check the version note."""

        version = parse_tag("10.1.0-beta1")
        self.assertIsInstance(version, VersionNote)
        self.assertEqual("10", version.major)
        self.assertEqual("1", version.minor)
        self.assertEqual("0", version.patch)
        self.assertEqual("beta1", version.prerelease)
        self.assertTrue(version.is_prerelease)

        version = parse_tag("3.4.0-rc.2")
        self.assertIsInstance(version, VersionNote)
        self.assertEqual("3", version.major)
        self.assertEqual("4", version.minor)
        self.assertEqual("0", version.patch)
        self.assertEqual("rc.2", version.prerelease)
        self.assertTrue(version.is_prerelease)

        version = parse_tag("10.1.0")
        self.assertIsInstance(version, VersionNote)
        self.assertEqual("10", version.major)
        self.assertEqual("1", version.minor)
        self.assertEqual("0", version.patch)
        self.assertIsNone(version.prerelease)
        self.assertFalse(version.is_prerelease)

        version = parse_tag("v10.1.0")
        self.assertIsInstance(version, VersionNote)
        self.assertEqual("v10", version.major)
        self.assertEqual("1", version.minor)
        self.assertEqual("0", version.patch)
        self.assertIsNone(version.prerelease)
        self.assertFalse(version.is_prerelease)

        # Not following https://semver.org/, we can't guess
        version = parse_tag("10.1")
        self.assertIsInstance(version, VersionNote)
        self.assertIsNone(version.major)
        self.assertIsNone(version.minor)
        self.assertIsNone(version.patch)
        self.assertIsNone(version.prerelease)
        self.assertFalse(version.is_prerelease)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3

import logging
import os
import re
import sys
import tarfile
import xmlrpc.client
import zipfile
from glob import glob
from pathlib import Path
from tempfile import mkstemp

import git
from github import Github, GithubException

try:
    import importlib.resources as importlib_resources
except ImportError:
    # In Py<3.7 fall-back to backported `importlib_resources`.
    import importlib_resources

import datetime

import pyqt5ac

from qgispluginci.changelog import ChangelogParser
from qgispluginci.exceptions import (
    BuiltResourceInSources,
    GithubReleaseCouldNotUploadAsset,
    GithubReleaseNotFound,
    MissingChangelog,
    UncommitedChanges,
)
from qgispluginci.parameters import Parameters
from qgispluginci.translation import Translation
from qgispluginci.utils import (
    configure_file,
    convert_octets,
    parse_tag,
    replace_in_file,
)

# GLOBALS
logger = logging.getLogger(__name__)


def create_archive(
    parameters: Parameters,
    release_version: str,
    archive_name: str,
    add_translations: bool = False,
    allow_uncommitted_changes: bool = False,
    is_prerelease: bool = False,
    raise_min_version: str = None,
    disable_submodule_update: bool = False,
):

    repo = git.Repo()

    top_tar_handle, top_tar_file = mkstemp(suffix=".tar")

    # keep track of current state
    initial_stash = None
    diff = repo.index.diff(None)
    if diff:
        logger.info("There are uncommitted changes:")
        for diff in diff:
            logger.info(diff)
        if not allow_uncommitted_changes:
            err_msg = (
                "You have uncommitted changes. "
                "Stash or commit them or use -c / --allow-uncommitted-changes option."
            )
            logger.error(err_msg, exc_info=UncommitedChanges())
            sys.exit(1)
        else:
            initial_stash = repo.git.stash("create")

    # changelog
    if parameters.changelog_include:
        parser = ChangelogParser(
            parent_folder=Path(parameters.plugin_path).resolve().parent,
            changelog_path=parameters.changelog_path,
        )
        if parser.has_changelog():
            try:
                content = parser.last_items(
                    count=parameters.changelog_number_of_entries
                )
                if content:
                    replace_in_file(
                        "{}/metadata.txt".format(parameters.plugin_path),
                        r"^changelog=.*$",
                        "changelog={}".format(content),
                    )
            except Exception as exc:
                # Do not fail the release process if something is wrong when parsing the changelog
                replace_in_file(
                    "{}/metadata.txt".format(parameters.plugin_path),
                    r"^changelog=.*$",
                    "",
                )
                logger.warning(
                    f"An exception occurred while parsing the changelog file: {exc}",
                    exc_info=exc,
                )
    else:
        # Remove the changelog line
        replace_in_file(
            "{}/metadata.txt".format(parameters.plugin_path), r"^changelog=.*$", ""
        )

    # set version in metadata
    replace_in_file(
        "{}/metadata.txt".format(parameters.plugin_path),
        r"^version=.*$",
        "version={}".format(release_version),
    )

    # Commit number
    replace_in_file(
        f"{parameters.plugin_path}/metadata.txt",
        r"^commitNumber=.*$",
        f"commitNumber={len(list(repo.iter_commits()))}",
    )

    # Git SHA1
    replace_in_file(
        f"{parameters.plugin_path}/metadata.txt",
        r"^commitSha1=.*$",
        f"commitSha1={repo.head.object.hexsha}",
    )

    # Date/time in UTC
    date_time = datetime.datetime.now(datetime.timezone.utc).strftime(
        "%Y-%m-%dT%H:%M:%SZ"
    )
    replace_in_file(
        f"{parameters.plugin_path}/metadata.txt",
        r"^dateTime=.*$",
        f"dateTime={date_time}",
    )

    # set the plugin as experimental on a pre-release
    if is_prerelease:
        replace_in_file(
            "{}/metadata.txt".format(parameters.plugin_path),
            r"^experimental=.*$",
            "experimental={}".format(True if is_prerelease else False),
        )

    if raise_min_version:
        replace_in_file(
            "{}/metadata.txt".format(parameters.plugin_path),
            r"^qgisMinimumVersion=.*$",
            "qgisMinimumVersion={}".format(raise_min_version),
        )

    # replace any DEBUG=False in all Python files
    if not is_prerelease:
        for file in glob("{}/**/*.py".format(parameters.plugin_path), recursive=True):
            replace_in_file(file, r"^DEBUG\s*=\s*True", "DEBUG = False")

    # keep track of current state
    try:
        stash = repo.git.stash("create")
    except git.exc.GitCommandError:
        stash = "HEAD"
    if stash == "" or stash is None:
        stash = "HEAD"
    # create TAR archive
    logger.debug(f"Git archive plugin with stash: {stash}")
    repo.git.archive(stash, "-o", top_tar_file, parameters.plugin_path)
    # adding submodules
    for submodule in repo.submodules:
        _, sub_tar_file = mkstemp(suffix=".tar")
        if submodule.path.split("/")[0] != parameters.plugin_path:
            logger.debug(
                f"Skipping submodule not in plugin source directory: {submodule.name}"
            )
            continue
        if not disable_submodule_update:
            submodule.update(init=True)
        sub_repo = submodule.module()
        logger.info("Git archive submodule: {sub_repo}")
        sub_repo.git.archive(
            "HEAD", "--prefix", "{}/".format(submodule.path), "-o", sub_tar_file
        )
        with tarfile.open(top_tar_file, mode="a") as tt:
            with tarfile.open(sub_tar_file, mode="r:") as st:
                for m in st.getmembers():
                    if not m.isfile():
                        continue
                    tt.add(m.name)

    # add translation files
    if add_translations:
        with tarfile.open(top_tar_file, mode="a") as tt:
            logger.debug("Adding translations")
            for file in glob("{}/i18n/*.qm".format(parameters.plugin_path)):
                logger.debug("  adding translation: {}".format(os.path.basename(file)))
                # https://stackoverflow.com/a/48462950/1548052
                tt.add(file)

    # compile qrc files
    if list(Path(parameters.plugin_path).glob("*.qrc")):
        pyqt5ac.main(
            ioPaths=[
                [
                    "{}/*.qrc".format(parameters.plugin_path),
                    "{}/%%FILENAME%%_rc.py".format(parameters.plugin_path),
                ]
            ]
        )
        for file in glob("{}/*_rc.py".format(parameters.plugin_path)):
            with tarfile.open(top_tar_file, mode="r:") as tt:
                for n in tt.getnames():
                    if n == file:
                        err_msg = (
                            f"The file {file} is present in the sources and its name "
                            "conflicts with a just built resource. "
                            "You might want to remove it from the sources or "
                            "setting export-ignore in .gitattributes config file."
                        )
                        logger.error(err_msg, exc_info=BuiltResourceInSources())
                        sys.exit(1)
            with tarfile.open(top_tar_file, mode="a") as tt:
                logger.debug("\tAdding resource: {}".format(file))
                # https://stackoverflow.com/a/48462950/1548052
                tt.add(file)

    # converting to ZIP
    # why using TAR before? because it provides the prefix and makes things easier
    with zipfile.ZipFile(
        file=archive_name, mode="w", compression=zipfile.ZIP_DEFLATED
    ) as zf:
        # adding the content of TAR archive
        with tarfile.open(top_tar_file, mode="r:") as tt:
            for m in tt.getmembers():
                if m.isdir():
                    continue
                f = tt.extractfile(m)
                fl = f.read()
                fn = m.name
                # Get permissions and add it to ZipInfo
                st = os.stat(m.name)
                info = zipfile.ZipInfo(fn)

                # Using flags as defined in python zipfile module
                # code : https://github.com/python/cpython/blob/b885b8f4be9c74ef1ce7923dbf055c31e7f47735/Lib/zipfile.py#L545
                # see https://stackoverflow.com/questions/434641/how-do-i-set-permissions-attributes-on-a-file-in-a-zip-file-using-pythons-zip/53008127#53008127
                info.external_attr = (st[0] & 0xFFFF) << 16  # Unix attributes
                info.compress_type = zf.compression
                zf.writestr(info, fl)

    logger.debug("-" * 40)
    logger.debug("Files in ZIP archive ({}):".format(archive_name))
    with zipfile.ZipFile(file=archive_name, mode="r") as zf:
        for f in zf.namelist():
            logger.debug(f)
    logger.debug("-" * 40)

    # checkout to reset changes
    if initial_stash:
        repo.git.reset("--hard", initial_stash)
        repo.git.reset("HEAD^")
    else:
        repo.git.checkout("--", ".")

    # print the result
    print(  # noqa: T2
        f"Plugin archive created: {archive_name} "
        f"({convert_octets(Path(archive_name).stat().st_size)})"
    )


def upload_asset_to_github_release(
    parameters: Parameters,
    asset_path: str,
    release_tag: str,
    github_token: str,
    asset_name: str = None,
):

    slug = "{}/{}".format(parameters.github_organization_slug, parameters.project_slug)
    repo = Github(github_token).get_repo(slug)
    try:
        logger.debug(
            f"Getting release on {parameters.github_organization_slug}"
            f"/{parameters.project_slug}"
        )
        gh_release = repo.get_release(id=release_tag)
        logger.debug(
            f"Release retrieved from GitHub: {gh_release}, "
            f"{gh_release.tag_name}, "
            f"{gh_release.upload_url}"
        )
    except GithubException as exc:
        logger.error(
            f"Release {release_tag} not found for {slug}",
            exc_info=GithubReleaseNotFound(exc),
        )
        sys.exit(1)
    try:
        assert os.path.exists(asset_path)
        if asset_name:
            logger.debug(f"Uploading asset: {asset_path} as {asset_name}")

            uploaded_asset = gh_release.upload_asset(
                path=asset_path, label=asset_name, name=asset_name
            )
        else:
            logger.debug(f"Uploading asset: {asset_path}")
            uploaded_asset = gh_release.upload_asset(asset_path)
        logger.info(f"Asset successfully uploaded: {uploaded_asset.url}")
    except GithubException as exc:
        logger.error(
            f"Could not upload asset for release {release_tag} on {slug}. "
            "Are you sure the user for the given token can upload asset to this repo?",
            exc_info=GithubReleaseCouldNotUploadAsset(exc),
        )
        sys.exit(1)


def release_is_prerelease(
    parameters: Parameters,
    release_tag: str,
    github_token: str,
) -> bool:
    """Check the tag name or the GitHub release if the version must be experimental or not."""

    if parse_tag(release_tag).is_prerelease:
        # The tag itself is a pre-release according to https://semver.org/
        return True

    if not github_token:
        return False

    slug = "{}/{}".format(parameters.github_organization_slug, parameters.project_slug)
    repo = Github(github_token).get_repo(slug)
    try:
        logger.debug(
            f"Getting release on {parameters.github_organization_slug}"
            f"/{parameters.project_slug}"
        )
        gh_release = repo.get_release(id=release_tag)
        logger.debug(
            f"Release retrieved from GitHub: {gh_release}, "
            f"{gh_release.tag_name}, "
            f"{gh_release.upload_url}"
        )
    except GithubException as exc:
        logger.error(
            f"Release {release_tag} not found. Trace: {exc}",
            exc_info=GithubReleaseNotFound(),
        )
        sys.exit(1)
    return gh_release.prerelease


def create_plugin_repo(
    parameters: Parameters,
    release_version: str,
    release_tag: str,
    is_prerelease: bool,
    archive: str,
    osgeo_username,
    plugin_repo_url=None,
) -> str:
    """
    Creates the plugin repo as an XML file
    """
    replace_dict = {
        "__RELEASE_VERSION__": release_version,
        "__RELEASE_TAG__": release_tag or release_version,
        "__PLUGIN_NAME__": parameters.plugin_name,
        "__RELEASE_DATE__": datetime.date.today().strftime("%Y-%m-%d"),
        "__CREATE_DATE__": parameters.create_date.strftime("%Y-%m-%d"),
        "__ORG__": parameters.github_organization_slug,
        "__REPO__": parameters.project_slug,
        "__PLUGINZIP__": archive,
        "__OSGEO_USERNAME__": osgeo_username or parameters.author,
        "__DEPRECATED__": str(parameters.deprecated),
        "__EXPERIMENTAL__": str(is_prerelease or parameters.experimental),
        "__TAGS__": parameters.tags,
        "__ICON__": parameters.icon,
        "__AUTHOR__": parameters.author,
        "__QGIS_MIN_VERSION__": parameters.qgis_minimum_version,
        "__DESCRIPTION__": parameters.description,
        "__ISSUE_TRACKER__": parameters.issue_tracker,
        "__HOMEPAGE__": parameters.homepage,
        "__REPO_URL__": parameters.repository_url,
    }
    if not plugin_repo_url:
        download_url = "https://github.com/{org}/{repo}/releases/download/{tag}/{pluginzip}".format(
            org=replace_dict["__ORG__"],
            repo=replace_dict["__REPO__"],
            tag=replace_dict["__RELEASE_TAG__"],
            pluginzip=replace_dict["__PLUGINZIP__"],
        )
        _, xml_repo = mkstemp(suffix=".xml")
    else:
        download_url = "{url}{pluginzip}".format(
            url=plugin_repo_url,
            pluginzip=replace_dict["__PLUGINZIP__"],
        )
        xml_repo = "./plugins.xml"
    replace_dict["__DOWNLOAD_URL__"] = download_url
    with importlib_resources.path(
        "qgispluginci", "plugins.xml.template"
    ) as xml_template:
        configure_file(xml_template, xml_repo, replace_dict)
    return xml_repo


def upload_plugin_to_osgeo(username: str, password: str, archive: str):
    """
    Upload the plugin to QGIS repository

    Parameters
    ----------
    username
        The username
    password
        The password
    archive
        The plugin archive file path to be uploaded
    """
    address = f"https://{username}:{password}@plugins.qgis.org:443/plugins/RPC2/"

    server = xmlrpc.client.ServerProxy(address, verbose=True)

    try:
        logger.debug(f"Start uploading {archive} to QGIS plugins repository.")
        with open(archive, "rb") as handle:
            plugin_id, version_id = server.plugin.upload(
                xmlrpc.client.Binary(handle.read())
            )
        logger.debug(f"Plugin ID: {plugin_id}")
        logger.debug(f"Version ID: {version_id}")
    except xmlrpc.client.ProtocolError as err:
        url = re.sub(r":[^/].*@", ":******@", err.url)
        err_msg = (
            "=== A protocol error occurred ===\n"
            f"URL: {url}\n"
            f"HTTP/HTTPS headers: {err.headers}\n"
            f"Error code: {err.errcode}\n"
            f"Error message: {err.errmsg}\n"
            f"Plugin path: {archive}"
        )
        logger.error(err_msg)
        sys.exit(err_msg)
    except xmlrpc.client.Fault as err:
        err_msg = (
            "=== A fault occurred occurred ===\n"
            f"Fault code: {err.faultCode}\n"
            f"Fault string: {err.faultString}\n"
            f"Plugin path: {archive}"
        )
        logger.error(err_msg)
        sys.exit(err_msg)


def release(
    parameters: Parameters,
    release_version: str,
    release_tag: str = None,
    github_token: str = None,
    upload_plugin_repo_github: bool = False,
    transifex_token: str = None,
    osgeo_username: str = None,
    osgeo_password: str = None,
    allow_uncommitted_changes: bool = False,
    plugin_repo_url: str = None,
    disable_submodule_update: bool = False,
):
    """

    Parameters
    ----------
    parameters
        The configuration parameters
    release_version:
        The release version (x.y.z)
    release_tag:
        The release tag (vx.y.z).
        If not given, the release version will be used
    github_token
        The Github token
    upload_plugin_repo_github
        If true, a custom repo will be created as a release asset on Github and could later be used in QGIS as a custom plugin repository.
    plugin_repo_url
        If set, this URL will be used to create the ZIP URL in the XML file
    transifex_token
        The Transifex token
    osgeo_username
        osgeo username to upload the plugin to official QGIS repository
    osgeo_password
        osgeo password to upload the plugin to official QGIS repository
    allow_uncommitted_changes
        If False, uncommitted changes are not allowed before packaging/releasing.
        If True and some changes are detected, a hard reset on a stash create will be used to revert changes made by qgis-plugin-ci.
    disable_submodule_update
        If omitted, a git submodule is updated. If specified, git submodules will not be updated/initialized before packaging.
    """

    if release_version == "latest":
        parser = ChangelogParser(
            parent_folder=Path(parameters.plugin_path).resolve().parent,
            changelog_path=parameters.changelog_path,
        )
        if parser.has_changelog():
            release_version = parser.latest_version()
        else:
            release_version = "latest"

    release_tag = release_tag or release_version

    if transifex_token is not None:
        tr = Translation(
            parameters, create_project=False, transifex_token=transifex_token
        )
        tr.pull()
        tr.compile_strings()

    archive_name = parameters.archive_name(parameters.plugin_path, release_version)

    # check if the GitHub release is a regular or pre-release
    is_prerelease = release_is_prerelease(
        parameters, release_tag=release_tag, github_token=github_token
    )

    if is_prerelease:
        logger.info(f"{release_tag} is a pre-release.")
    else:
        logger.info(f"{release_tag} is a regular release.")

    create_archive(
        parameters,
        release_version,
        archive_name,
        add_translations=transifex_token is not None,
        allow_uncommitted_changes=allow_uncommitted_changes,
        is_prerelease=is_prerelease,
        disable_submodule_update=disable_submodule_update,
    )

    # if pushing to QGIS repo and pre-release, create an extra package with qgisMinVersion to 3.14
    # since only QGIS 3.14+ supports the beta/experimental plugins trial
    experimental_archive_name = None
    if osgeo_username is not None and is_prerelease:
        experimental_archive_name = parameters.archive_name(
            parameters.plugin_path, release_version, True
        )
        create_archive(
            parameters,
            release_version,
            experimental_archive_name,
            add_translations=transifex_token is not None,
            allow_uncommitted_changes=allow_uncommitted_changes,
            is_prerelease=True,
            raise_min_version="3.14",
            disable_submodule_update=disable_submodule_update,
        )

    if github_token is not None:
        upload_asset_to_github_release(
            parameters,
            asset_path=archive_name,
            release_tag=release_tag,
            github_token=github_token,
        )
        if upload_plugin_repo_github:
            xml_repo = create_plugin_repo(
                parameters=parameters,
                release_version=release_version,
                release_tag=release_tag,
                is_prerelease=is_prerelease,
                archive=archive_name,
                osgeo_username=osgeo_username,
            )
            upload_asset_to_github_release(
                parameters,
                asset_path=xml_repo,
                release_tag=release_tag,
                github_token=github_token,
                asset_name="plugins.xml",
            )

    if plugin_repo_url:
        xml_repo = create_plugin_repo(
            parameters=parameters,
            release_version=release_version,
            release_tag=release_tag,
            archive=archive_name,
            osgeo_username=osgeo_username,
            plugin_repo_url=plugin_repo_url,
        )
        logger.info(f"Local XML repo file created : {xml_repo}")

    if osgeo_username is not None:
        assert osgeo_password is not None
        if is_prerelease:
            assert experimental_archive_name is not None
            upload_plugin_to_osgeo(
                username=osgeo_username,
                password=osgeo_password,
                archive=experimental_archive_name,
            )
        else:
            upload_plugin_to_osgeo(
                username=osgeo_username, password=osgeo_password, archive=archive_name
            )



import glob
import logging
import subprocess
import sys
from pathlib import Path

from pytransifex import Transifex

from qgispluginci.exceptions import (
    TransifexManyResources,
    TransifexNoResource,
    TranslationFailed,
)
from qgispluginci.parameters import Parameters
from qgispluginci.utils import touch_file

# GLOBALS
logger = logging.getLogger(__name__)


class Translation:
    def __init__(
        self, parameters: Parameters, transifex_token: str, create_project: bool = True
    ):
        """
        Parameters
        ----------
        parameters:

        transifex_token:
            Transifex API token

        create_project:
            if True, it will create the project, resource and language on Transifex

        """
        self.parameters = parameters
        self._t = Transifex(
            transifex_token, parameters.transifex_organization, i18n_type="QT"
        )
        assert self._t.ping()
        self.ts_file = "{dir}/i18n/{res}_{lan}.ts".format(
            dir=self.parameters.plugin_path,
            res=self.parameters.transifex_resource,
            lan=self.parameters.translation_source_language,
        )

        if self._t.project_exists(parameters.transifex_project):
            logger.debug(
                f"Project {self.parameters.transifex_organization}/"
                f"{self.parameters.transifex_project} exists on Transifex"
            )

        elif create_project:
            logger.debug(
                "Project does not exists on Transifex, creating one as: "
                f"{self.parameters.transifex_organization}/"
                f"{self.parameters.transifex_project}"
            )
            self._t.create_project(
                slug=self.parameters.transifex_project,
                repository_url=self.parameters.repository_url,
                source_language_code=parameters.translation_source_language,
            )
            self.update_strings()
            logger.debug(
                f"Creating resource in {self.parameters.transifex_organization}/"
                f"{self.parameters.transifex_project}/"
                f"{self.parameters.transifex_resource} with {self.ts_file}"
            )
            self._t.create_resource(
                project_slug=self.parameters.transifex_project,
                path_to_file=self.ts_file,
                resource_slug=self.parameters.transifex_resource,
            )
            logger.info(
                "Transifex project {self.parameters.transifex_organization}/"
                f"{self.parameters.transifex_project} and resource have been created."
            )
        else:
            logger.error(
                "Project does not exists on Transifex: "
                f"{self.parameters.transifex_organization}/"
                f"{self.parameters.transifex_project}",
                exc_info=TranslationFailed(),
            )

    def update_strings(self):
        """
        Update TS files from plugin source strings
        """
        source_py_files = []
        source_ui_files = []
        relative_path = "./{plugin_path}".format(
            plugin_path=self.parameters.plugin_path
        )
        for ext in ("py", "ui"):
            for file in glob.glob(
                "{dir}/**/*.{ext}".format(dir=self.parameters.plugin_path, ext=ext),
                recursive=True,
            ):
                file_path = str(Path(file).relative_to(relative_path))
                if ext == "py":
                    source_py_files.append(file_path)
                else:
                    source_ui_files.append(file_path)

        touch_file(self.ts_file)

        project_file = Path(self.parameters.plugin_path).joinpath(
            self.parameters.plugin_name + ".pro"
        )

        with open(project_file, "w") as f:
            assert f.write("CODECFORTR = UTF-8\n")
            assert f.write("SOURCES = {}\n".format(" ".join(source_py_files)))
            assert f.write("FORMS = {}\n".format(" ".join(source_ui_files)))
            assert f.write(
                "TRANSLATIONS = {}\n".format(
                    Path(self.ts_file).relative_to(relative_path)
                )
            )
            f.flush()
            f.close()

        cmd = [self.parameters.pylupdate5_path, "-noobsolete", str(project_file)]

        output = subprocess.run(cmd, capture_output=True, text=True)

        project_file.unlink()

        if output.returncode != 0:
            logger.error(
                f"Translation failed: {output.stderr}", exc_info=TranslationFailed()
            )
            sys.exit(1)
        else:
            logger.info(f"Successfully run pylupdate5: {output.stdout}")

    def compile_strings(self):
        """
        Compile TS file into QM files
        """
        cmd = [self.parameters.lrelease_path]
        for file in glob.glob(
            "{dir}/i18n/*.ts".format(dir=self.parameters.plugin_path)
        ):
            cmd.append(file)
        output = subprocess.run(cmd, capture_output=True, text=True)
        if output.returncode != 0:
            logger.error(
                f"Translation failed: {output.stderr}", exc_info=TranslationFailed()
            )
            sys.exit(1)
        else:
            logger.info(f"Successfully run lrelease: {output.stdout}")

    def pull(self):
        """
        Pull TS files from Transifex
        """
        resource = self.__get_resource()
        existing_langs = self._t.list_languages(
            project_slug=self.parameters.transifex_project,
            resource_slug=resource["slug"],
        )
        existing_langs.remove(self.parameters.translation_source_language)
        logger.info(
            f"{len(existing_langs)} languages found for resource {resource.get('slug')}:"
            f" ({existing_langs})"
        )
        for lang in self.parameters.translation_languages:
            if lang not in existing_langs:
                logger.debug(f"Creating missing language: {lang}")
                self._t.create_language(
                    self.parameters.transifex_project,
                    lang,
                    [self.parameters.transifex_coordinator],
                )
                existing_langs.append(lang)
        for lang in existing_langs:
            ts_file = "{dir}/i18n/{res}_{lan}.ts".format(
                dir=self.parameters.plugin_path,
                res=self.parameters.transifex_resource,
                lan=lang,
            )
            logger.debug(f"Downloading translation file: {ts_file}")
            self._t.get_translation(
                self.parameters.transifex_project, resource["slug"], lang, ts_file
            )

    def push(self):
        resource = self.__get_resource()
        logger.debug(
            f"Pushing resource: {self.parameters.transifex_resource} "
            f"with file {self.ts_file}"
        )
        result = self._t.update_source_translation(
            project_slug=self.parameters.transifex_project,
            resource_slug=resource["slug"],
            path_to_file=self.ts_file,
        )
        logger.info(f"Translation resource updated: {result}")

    def __get_resource(self) -> dict:
        resources = self._t.list_resources(self.parameters.transifex_project)
        if len(resources) == 0:
            logger.error(
                f"Project '{self.parameters.transifex_project}' has no resource on Transifex",
                exc_info=TransifexNoResource(),
            )
            sys.exit(1)
        if len(resources) > 1:
            for resource in resources:
                if resource["name"] == self.parameters.transifex_resource:
                    return resource
            logger.error(
                f"Project '{self.parameters.transifex_project}' has several "
                "resources on Transifex and none is named as the project slug. "
                "Specify one in the parameters with transifex_resource."
                "These resources have been found: "
                f"{', '.join([r['name'] for r in resources])}",
                exc_info=TransifexManyResources(),
            )
            sys.exit(1)
        return resources[0]



#! python3  # noqa E265

"""
    Parameters management.
"""

# ############################################################################
# ########## Libraries #############
# ##################################

# standard library
import datetime
import logging
import os
import re
import sys

# 3rd party
from slugify import slugify

# ############################################################################
# ########## Globals #############
# ################################


DASH_WARNING = "Dash in the plugin name is causing issues with QGIS plugin manager"

logger = logging.getLogger(__name__)


# ############################################################################
# ########## Classes #############
# ################################


class Parameters:
    """
    Attributes
    ----------
    plugin_path: str
        The directory of the source code in the repository.
        Defaults to: `slugify(plugin_name)`

    github_organization_slug: str
        The organization slug on SCM host (e.g. Github) and translation platform (e.g. Transifex).
        Not required when running on Travis since deduced from `$TRAVIS_REPO_SLUG`environment variable.

    project_slug: str
        The project slug on SCM host (e.g. Github) and translation platform (e.g. Transifex).
        Not required when running on Travis since deduced from `$TRAVIS_REPO_SLUG`environment variable.

    transifex_coordinator: str
        The username of the coordinator in Transifex.
        Required to create new languages.

    transifex_organization: str
        The organization name in Transifex
        Defaults to: the GitHub organization slug

    transifex_project: str
        The project on Transifex, which can be different from the one on GitHub.
        Defaults to: the project_slug

    transifex_resource: str
        The resource name in transifex
        Defaults to: the project_slug

    translation_source_language:
        The source language for translations.
        Defaults to: 'en'

    translation_languages:
        List of languages.

    changelog_include:
        If the changelog must be added when releasing a version AND if there is a CHANGELOG.md file
        Defaults to True

    changelog_path:
        Path to the CHANGELOG.md relative to the configuration file. Defaults to CHANGELOG.md

    changelog_number_of_entries:
        Number of changelog entries to add in the metdata.txt
        Defaults to 3

    create_date: datetime.date
        The date of creation of the plugin.
        The would be used in the custom repository XML.
        Format: YYYY-MM-DD

    lrelease_path: str
        The path of lrelease executable

    pylupdate5_path: str
        The path of pylupdate executable


    """

    def __init__(self, definition: dict):
        self.plugin_path = definition.get("plugin_path")
        self.plugin_name = self.__get_from_metadata("name")
        self.plugin_slug = slugify(self.plugin_name)
        self.project_slug = definition.get(
            "project_slug",
            os.environ.get("TRAVIS_REPO_SLUG", ".../{}".format(self.plugin_slug)).split(
                "/"
            )[1],
        )
        self.github_organization_slug = definition.get(
            "github_organization_slug",
            os.environ.get("TRAVIS_REPO_SLUG", "").split("/")[0],
        )
        self.transifex_coordinator = definition.get("transifex_coordinator", "")
        self.transifex_organization = definition.get(
            "transifex_organization", self.github_organization_slug
        )
        self.translation_source_language = definition.get(
            "translation_source_language", "en"
        )
        self.translation_languages = definition.get("translation_languages", {})
        self.transifex_project = definition.get("transifex_project", self.project_slug)
        self.transifex_resource = definition.get(
            "transifex_resource", self.project_slug
        )
        self.create_date = datetime.datetime.strptime(
            str(definition.get("create_date", datetime.date.today())), "%Y-%m-%d"
        )
        self.lrelease_path = definition.get("lrelease_path", "lrelease")
        self.pylupdate5_path = definition.get("pylupdate5_path", "pylupdate5")
        changelog_include = definition.get("changelog_include", True)
        if isinstance(changelog_include, str):
            self.changelog_include = changelog_include.lower() in [
                "true",
                "1",
                "t",
                "y",
            ]
        else:
            self.changelog_include = changelog_include
        self.changelog_number_of_entries = definition.get(
            "changelog_number_of_entries", 3
        )
        self.changelog_path = definition.get("changelog_path", "CHANGELOG.md")

        # read from metadata
        if not self.plugin_path:
            # This tool can be used outside of a QGIS plugin to read a changelog file
            return

        self.author = self.__get_from_metadata("author", "")
        self.description = self.__get_from_metadata("description")
        self.qgis_minimum_version = self.__get_from_metadata("qgisMinimumVersion")
        self.icon = self.__get_from_metadata("icon", "")
        self.tags = self.__get_from_metadata("tags", "")
        self.experimental = self.__get_from_metadata("experimental", False)
        self.deprecated = self.__get_from_metadata("deprecated", False)
        self.issue_tracker = self.__get_from_metadata("tracker")
        self.homepage = self.__get_from_metadata("homepage", "")
        if self.homepage == "":
            logger.warning(
                "Homepage is not given in the metadata. "
                "It is a mandatory information to publish "
                "the plugin on the QGIS official repository."
            )
        self.repository_url = self.__get_from_metadata("repository")

    @staticmethod
    def archive_name(
        plugin_name, release_version: str, experimental: bool = False
    ) -> str:
        """
        Returns the archive file name
        """
        # zipname: use dot before version number
        # and not dash since it's causing issues #22
        if "-" in plugin_name:
            logger.warning(DASH_WARNING)

        return "{zipname}{experimental}.{release_version}.zip".format(
            zipname=plugin_name,
            experimental="-experimental" if experimental else "",
            release_version=release_version,
        )

    def __get_from_metadata(self, key: str, default_value: any = None) -> str:
        if not self.plugin_path:
            return ""

        metadata_file = f"{self.plugin_path}/metadata.txt"
        with open(metadata_file) as f:
            for line in f:
                m = re.match(r"{}\s*=\s*(.*)$".format(key), line)
                if m:
                    return m.group(1)
        if default_value is None:
            logger.error(f"Mandatory key is missing in metadata: {key}")
            sys.exit(1)
        return default_value


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    pass



#! /usr/bin/env python

# standard
import filecmp
import os
import re
import unittest
import urllib.request
from pathlib import Path
from tempfile import mkstemp
from zipfile import ZipFile

# 3rd party
import yaml
from github import Github, GithubException
from pytransifex.exceptions import PyTransifexException

# Project
from qgispluginci.changelog import ChangelogParser
from qgispluginci.exceptions import GithubReleaseNotFound
from qgispluginci.parameters import DASH_WARNING, Parameters
from qgispluginci.release import release
from qgispluginci.translation import Translation
from qgispluginci.utils import replace_in_file

# Tests
from .utils import can_skip_test

# If changed, also update CHANGELOG.md
RELEASE_VERSION_TEST = "0.1.2"


class TestRelease(unittest.TestCase):
    def setUp(self):
        with open(".qgis-plugin-ci", "r", encoding="utf8") as f:
            arg_dict = yaml.safe_load(f)
        self.parameters = Parameters(arg_dict)
        self.transifex_token = os.getenv("transifex_token")
        self.github_token = os.getenv("github_token")
        self.repo = None
        self.t = None
        if self.github_token:
            print("init Github")
            self.repo = Github(self.github_token).get_repo("opengisch/qgis-plugin-ci")
        self.clean_assets()

    def tearDown(self):
        self.clean_assets()

    def clean_assets(self):
        if self.repo:
            rel = None
            try:
                rel = self.repo.get_release(id=RELEASE_VERSION_TEST)
            except GithubException:
                raise GithubReleaseNotFound(
                    "Release {} not found".format(RELEASE_VERSION_TEST)
                )
            if rel:
                print("deleting release assets")
                for asset in rel.get_assets():
                    print("  delete {}".format(asset.name))
                    asset.delete_asset()
        if self.t:
            try:
                self.t._t.delete_project(self.parameters.project_slug)
            except PyTransifexException:
                pass
            try:
                self.t._t.delete_team("{}-team".format(self.parameters.project_slug))
            except PyTransifexException:
                pass

    def test_release(self):
        release(self.parameters, RELEASE_VERSION_TEST)

    @unittest.skipIf(can_skip_test(), "Missing transifex_token")
    def test_release_with_transifex(self):
        Translation(self.parameters, transifex_token=self.transifex_token)
        release(
            self.parameters, RELEASE_VERSION_TEST, transifex_token=self.transifex_token
        )

    def test_zipname(self):
        """Tests about the zipname for the QGIS plugin manager.

        See #22 about dash
        and also capital letters
        """
        self.assertEqual(
            "my_plugin-experimental.0.0.0.zip",
            Parameters.archive_name("my_plugin", "0.0.0", True),
        )

        self.assertEqual(
            "My_Plugin.0.0.0.zip", Parameters.archive_name("My_Plugin", "0.0.0", False)
        )

        with self.assertLogs(
            logger="qgispluginci.parameters", level="WARNING"
        ) as captured:
            Parameters.archive_name("my-plugin", "0.0.0")
        self.assertEqual(
            len(captured.records), 1
        )  # check that there is only one log message
        self.assertEqual(captured.records[0].getMessage(), DASH_WARNING)

    @unittest.skipIf(can_skip_test(), "Missing github_token")
    def test_release_upload_github(self):
        release(
            self.parameters,
            RELEASE_VERSION_TEST,
            github_token=self.github_token,
            upload_plugin_repo_github=True,
        )

        # check the custom plugin repo
        _, xml_repo = mkstemp(suffix=".xml")
        url = "https://github.com/opengisch/qgis-plugin-ci/releases/download/{}/plugins.xml".format(
            RELEASE_VERSION_TEST
        )
        print("retrieve repo from {}".format(url))
        urllib.request.urlretrieve(url, xml_repo)
        replace_in_file(
            xml_repo,
            r"<update_date>[\w-]+<\/update_date>",
            "<update_date>__TODAY__</update_date>",
        )
        if not filecmp.cmp("test/plugins.xml.expected", xml_repo, shallow=False):
            import difflib

            with open("test/plugins.xml.expected") as f:
                text1 = f.readlines()
            with open(xml_repo) as f:
                text2 = f.readlines()
            self.assertFalse(True, "\n".join(difflib.unified_diff(text1, text2)))

        # compare archive file size
        gh_release = self.repo.get_release(id=RELEASE_VERSION_TEST)
        archive_name = self.parameters.archive_name(
            self.parameters.plugin_path, RELEASE_VERSION_TEST
        )
        fs = os.path.getsize(archive_name)
        print("size: ", fs)
        self.assertGreater(fs, 0, "archive file size must be > 0")
        found = False
        for a in gh_release.get_assets():
            if a.name == archive_name:
                found = True
                self.assertEqual(fs, a.size, "asset size doesn't march archive size.")
                break
        self.assertTrue(found, "asset not found")

    def test_release_changelog(self):
        """Test if changelog in metadata.txt inside zipped plugin after release command."""
        # variables
        cli_config_changelog = Path("test/fixtures/.qgis-plugin-ci-test-changelog.yaml")
        version_to_release = "0.1.2"

        # load specific parameters
        with cli_config_changelog.open() as in_cfg:
            arg_dict = yaml.safe_load(in_cfg)
        parameters = Parameters(arg_dict)
        self.assertIsInstance(parameters, Parameters)

        # get output zip path
        archive_name = parameters.archive_name(
            plugin_name=parameters.plugin_path, release_version=version_to_release
        )

        # extract last items from changelog
        parser = ChangelogParser()
        self.assertTrue(parser.has_changelog())
        changelog_lastitems = parser.last_items(
            count=parameters.changelog_number_of_entries
        )

        # Include a changelog
        release(
            parameters=parameters,
            release_version=version_to_release,
            allow_uncommitted_changes=True,
        )

        # open archive and compare
        with ZipFile(archive_name, "r") as zip_file:
            data = zip_file.read(f"{parameters.plugin_path}/metadata.txt")

        # Changelog
        self.assertGreater(
            data.find(bytes(changelog_lastitems, "utf8")),
            0,
            f"changelog detection failed in release: {data}",
        )

        # Commit number
        self.assertEqual(1, len(re.findall(r"commitNumber=\d+", str(data))))

        # Commit sha1 not in the metadata.txt
        self.assertEqual(0, len(re.findall(r"commitSha1=\d+", str(data))))


if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

# ############################################################################
# ########## Libraries #############
# ##################################

# standard library
import pathlib
import sys

# 3rd party
from setuptools import setup

# package (to get version)
from qgispluginci import __about__

# ############################################################################
# ########## Globals #############
# ################################

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
python_min_version = (3, 7)

# This string might be updated on CI on runtime with a proper semantic version name with X.Y.Z
VERSION = "2.5.0"

if "." not in VERSION:
    # If VERSION is still not a proper semantic versioning with X.Y.Z
    # let's hardcode 0.0.0
    VERSION = "0.0.0"

# ############################################################################
# ########## Setup #############
# ##############################


if sys.version_info < python_min_version:
    sys.exit(
        "qgis-plugin-ci requires at least Python version {vmaj}.{vmin}.\n"
        "You are currently running this installation with\n\n{curver}".format(
            vmaj=python_min_version[0], vmin=python_min_version[1], curver=sys.version
        )
    )

setup(
    name="qgis-plugin-ci",
    author=__about__.__author__,
    author_email=__about__.__email__,
    description=__about__.__summary__,
    packages=["qgispluginci", "scripts"],
    long_description=README,
    long_description_content_type="text/markdown",
    entry_points={"console_scripts": ["qgis-plugin-ci = scripts.qgis_plugin_ci:main"]},
    package_data={"qgispluginci": ["plugins.xml.template"]},
    version=VERSION,
    url=__about__.__uri__,
    project_urls={
        "Docs": "https://opengisch.github.io/qgis-plugin-ci/",
        "Bug Reports": "{}issues/".format(__about__.__uri__),
        "Source": __about__.__uri__,
    },
    download_url="https://github.com/opengisch/qgis-plugin-ci/archive/{}.tar.gz".format(
        VERSION
    ),
    install_requires=[
        "GitPython>=3.1,<3.2",
        "PyGithub>=1.54,<1.56",
        "PyQt5>=5.15,<5.16",
        "pyqt5ac>=1.2,<1.3",
        "python-slugify>=4.0,<4.1",
        "pytransifex>=0.1.10,<0.2",
        "pyyaml>=5.4,<5.5",
    ],
    extras_require={
        "dev": ["black", "flake8", "pre-commit"],
    },
    python_requires=">={vmaj}.{vmin}".format(
        vmaj=python_min_version[0], vmin=python_min_version[1]
    ),
    # metadata
    keywords=["QGIS", "CI", "changelog", "plugin"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: GIS",
    ],
)



class BuiltResourceInSources(Exception):
    pass


class ConfigurationNotFound(Exception):
    pass


class GithubReleaseCouldNotUploadAsset(Exception):
    pass


class GithubReleaseNotFound(Exception):
    pass


class MissingChangelog(Exception):
    pass


class TransifexManyResources(Warning):
    pass


class TransifexNoResource(Exception):
    pass


class TranslationFailed(Exception):
    pass


class UncommitedChanges(Exception):
    pass



#! python3  # noqa: E265

"""
    Metadata about the package to easily retrieve informations about it.

    See: https://packaging.python.org/guides/single-sourcing-package-version/
"""

# ############################################################################
# ########## Libraries #############
# ##################################

# standard library
from datetime import date

# ############################################################################
# ########## Globals #############
# ################################
__all__ = [
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__title_clean__",
    "__uri__",
    "__version__",
    "__version_info__",
]

__author__ = "Denis Rouzaud, Étienne Trimaille, Julien Moura"
__copyright__ = "2019 - {0}, {1}".format(date.today().year, __author__)
__email__ = "denis.rouzaud@gmail.com"
__license__ = "GNU General Public License v3.0"
__summary__ = (
    "Let qgis-plugin-ci package and release your QGIS plugins for you. "
    "Have a tea or go hiking meanwhile."
    "Contains scripts to perform automated testing and deployment for QGIS plugins. "
    "These scripts are written for and tested on GitHub, Travis-CI, GitHub workflows and Transifex."
)
__title__ = "QGIS Plugin CI"
__title_clean__ = "".join(e for e in __title__ if e.isalnum())
__uri__ = "https://github.com/opengisch/qgis-plugin-ci/"

# This string might be updated on CI on runtime with a proper semantic version name with X.Y.Z
__version__ = "2.5.0"

if "." not in __version__:
    # If __version__ is still not a proper semantic versioning with X.Y.Z
    # let's hardcode 0.0.0
    __version__ = "0.0.0"

__version_info__ = tuple(
    [
        int(num) if num.isdigit() else num
        for num in __version__.replace("-", ".", 1).split(".")
    ]
)
